/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2002 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   IO architecture. Base classes and Mixer implementation.
 */

#ifndef GSML_IO_H
#define GSML_IO_H

#include "Threads.hh" // Thread / Job support

#include "include.hh"
#include "Basics.hh"
#include "typedefs.hh"
#include "SampleBuffer.hh"

#include <vector>
#include <deque>

// #define DEBUG_GSML_IO
// #define DEBUG_GSML_ALSA

namespace GSML
{
  /* Documentation / Brainstorming ...
   * Must be able to:
   *  - 1. Normal IO mode
   *  - 2. Realtime Preview mode
   *  - 3. disk writer
   */
  
  /* The mixer is used to remix the track for playback or export into
   * a file. The node's channel's are assigned 1:1 to the MixerSource
   * tracks - Non existing tracks are simply "muted" - missing tracks
   * are also filled with zero data.
   */
  
  /* All major task are done via this Mixer class. The other class are
   * used for the input and output on the mixer's _sides_ and to
   * controll the mixing process. The params are read on io start and
   * only updated on a signal - so we can optimize the mixing loop by
   * _caching_ the mixer values/params.
   */
  
  /* There are additinal MixerSource classes that can be stacked upon
   * each other and act as filters (mainly) for the DataLayer
   * source. That way we implement looping and Markers (limmiting and
   * modifing of the start, end and current position ...):
   *
   *  DataSource <-> QueueSource <-> SelectionFilterSource <->
   *  (MultiplexerSourece) <-> MarkerFilterSource <-> LoopFilterSource
   *  <-> Mixer <-> MixerNode
   *
   * The QueueSource implements read-ahead and write-buffering to
   * decouple the Sound I/O from the blocking disk access (espcially
   * noticed on NFS mounted filesystems ...).
   *
   * A MultiplexerSource suppots switching the SourceChain. This is
   * mainly used to switch to a FX prehearsal stream.
   */
  
  /* They must have seperate handling for each track's io position
   * (easier handling for non-iterleaved disk writers). And a general
   * _master_ io position which is returned with the
   * Current-Read/Write-Pos (for the mixer to get the position after
   * unmuting tracks ...). The master pos is changed via the
   * Read/Write- methods by setting the master flag. (The track must
   * be a vaild track, the track and the dest _should_ be zero, and NO
   * data must be processed.)
   */
  
  using namespace std;
  using namespace SigC;
  
  class Mixer; // forward
  
  // the source (DataLayer or FX) to get data from (or capture to)
  class MixerSource : public SigC::Object
  {
  public:
    class Except : public BasicExcept {};
    class ExceptParamsInvaild : public Except {};
    class ExceptNotInputable : public Except {};
    class ExceptIORunning : public Except {};
    class ExceptNoIORunning : public Except {};
    class ExceptTrackInvaild : public Except {};
    
    MixerSource ();
    virtual ~MixerSource ();
    
    virtual track_index PlaybackTracks () const;
    virtual track_index CaptureTracks () const;
    
    virtual sample_index MinStart () const = 0;
    virtual sample_index MaxEnd () const = 0;
    
    // prepare Input fot the stream - might throw a params-invaild except
    virtual void StartO ();
    virtual void StartIO ();
    virtual void StopIO ();
    
    virtual bool IsIORunning () const;
    virtual bool IsInputable () const = 0;
    
    virtual sample_index CurrentReadPos () const;
    virtual sample_index CurrentWritePos () const;
    
    virtual void SetReadPos (track_index track, sample_index n_pos, bool is_master = false);
    virtual void SetWritePos (track_index track, sample_index n_pos, bool is_master = false);
    
    virtual bool Read (track_index track, sample_index length, SampleBufferDefault* dest,
		       bool is_master = false);
    virtual bool Write (track_index track, sample_index length, SampleBufferDefault* src,
			sample_index buffer_index, bool is_master = false);
    
    virtual void Debug () const;
    
    // for signalling the layers above that the start/end pos changed (looping ...)
    Signal1 <void, sample_index> SignalMinStartChanged;
    Signal1 <void, sample_index> SignalMaxEndChanged;
    
  protected:
    
    sample_index& CurrentReadPosition (track_index track, bool is_master) {
      if (is_master)
	return master_read_position;
      return io_read_positions [track];
    }
    
    sample_index&  CurrentWritePosition (track_index track, bool is_master) {
      if (is_master)
	return master_write_position;
      return io_write_positions [track];
    }
    
    void BuildTables (track_index n_start);
    void ClearTables ();
    
    virtual void StartO_impl () = 0;
    virtual void StartIO_impl () = 0;
    virtual void StopIO_impl () = 0;
    
    virtual void SetReadPos_impl (track_index track, sample_index n_pos, bool is_master) = 0;
    virtual void SetWritePos_impl (track_index track, sample_index n_pos, bool is_master) = 0;
    
    virtual bool Read_impl (track_index track, sample_index length, SampleBufferDefault* dest,
			    bool is_master) = 0;
    virtual bool Write_impl (track_index track, sample_index length, SampleBufferDefault* src,
			     sample_index buffer_index, bool is_master) = 0;
    virtual void Debug_impl () const = 0;
    
    // internal states
    bool io_running;
    bool io_input_running;
    
    track_index m_playback_tracks;
    track_index m_capture_tracks;
    
    sample_index master_read_position;
    sample_index master_write_position;
    
    vector <sample_index> io_read_positions;
    vector <sample_index> io_write_positions;
  };
  
  // a class that provides the interface for the mixer values
  class MixerParams
  {
  public:
    class Except : public BasicExcept {};
    class ExceptTrackInvaild : public Except {};
    class ExceptChannelInvaild : public Except {};
    class ExceptIORunning : public Except {};
    
    virtual ~MixerParams () {};
    
    virtual void SetTracks (track_index n_tracks) = 0;
    virtual track_index Tracks () = 0;
    
    virtual void SetChannels (track_index n_channels) = 0;
    virtual track_index Channels () = 0;
    
    virtual void SetTrackMute (track_index track, bool n_state) = 0;
    virtual bool IsTrackMuted (track_index track) = 0;
    
    virtual void SetChannelMute (track_index channel, bool n_state) = 0;
    virtual bool IsChannelMuted (track_index channel) = 0;
    
    virtual pcm_default::value_type Value (track_index track, track_index channel) = 0;
    
    // lock during IO - forbids track / channel changes (wouldn't be that nice for the mixer ...)
    virtual void StartIO () = 0;
    virtual void StopIO () = 0;
    virtual bool IsIORunning () = 0;
    
    virtual void Debug () = 0;
    
    // for other to realtime display (or render current playback /caputure) them ;-)
    Signal1<void, track_index> SignalTracksChanged;
    Signal1<void, track_index> SignalChannelsChanged;
    
    Signal2<void, track_index, bool> SignalTrackMuteChanged;
    Signal2<void, track_index, bool> SignalChannelMuteChanged;
    
    Signal3<void, track_index, track_index, pcm_default::value_type> SignalValueChanged;
  };
  
  // a node (IO implementation or disk writer) to stream the data to (or capture from)
  class MixerNode
  {
  public:
    class Except : public BasicExcept {};
    class ExceptNotInitialized : public Except {};
    class ExceptInitialized : public Except {};
    class ExceptHWError : public Except  // some basic HW interface problem
    {
    public:
      ExceptHWError (const char* i)
      {
	info = i;
      }
    };
    
    class ExceptIORunning : public Except {};
    class ExceptNoIORunning : public Except {};
    
    virtual ~MixerNode () {};
    
    virtual track_index PlaybackChannels () = 0;
    virtual track_index CaptureChannels () = 0;
    
    virtual int PlaybackPeriodSize () = 0;
    virtual int CapturePeriodSize () = 0;
    
    virtual int PlaybackPeriods () = 0;
    virtual int CapturePeriods () = 0;
    
    virtual int PlaybackXRuns () = 0;
    virtual int CaptureXRuns () = 0;
    
    virtual void SetPlayback (bool state) = 0;
    virtual void SetCapture (bool state) = 0;
    
    virtual bool IsPlayback () = 0;
    virtual bool IsCapture () = 0;
    
    virtual void StartIO (Mixer& source) = 0; // gets a prepared sources!
    virtual void StopIO () = 0;  // mustn't fixate the source! -> call StopFinishedNotification
    
    virtual bool IsIORunning () = 0;
    
    virtual sample_index GetCurrentPos () = 0;
    
    virtual void Debug () = 0;
  };
  
  // Additional stuff for a real hw (or netwoek) realtime interface
  class HWMixerNode : public MixerNode
  {
  public:
    
    struct Params
    {
      sample_rate min_rate;
      sample_rate max_rate;
      sample_rate cur_rate;
      
      int min_depth;
      int max_depth;
      int cur_depth;
      
      int min_period_time;
      int max_period_time;
      int cur_period_time;
      
      int min_periods;
      int max_periods;
      int cur_periods;
    };
    
    virtual ~HWMixerNode () {};
    // must be called after construcing the class - to do some init stuff (search cards ...)
    virtual bool Init () = 0;
    
    virtual const vector<string>& Interfaces () = 0;
    virtual const string& Interface () = 0;
    virtual bool SetInterface (const string& n_interface) = 0;
    
    virtual const Params& GetParams ();
   
    virtual sample_rate SetRate (sample_rate n_rate);
    virtual int SetDepth (int n_depth);
    virtual int SetPeriodTime (int n_period_time);
    virtual int SetPeriods (int n_periods);
  
  protected:
    vector<string> m_interfaces;
    Params m_params;
  };
  
  // the mixer
  class Mixer : public SigC::Object
  {
  public:
    
    class Except : public BasicExcept {};
    class ExceptIORunning : public Except {};
    class ExceptNoIORunning : public Except {};
    class ExceptSourceBusy : public Except {};
    class ExceptNodeBusy : public Except {};
    class ExceptNoIOSelected : public Except {};
    class ExceptNoNode : public Except {};
    
    struct vu_meter_data
    {
      pcm_default::value_type GetMaxAbs ()
      {
	pcm_default::value_type t_data;
	t_data (limit_min (abs (min.data), abs (max.data) ) );
        return t_data;
      }
      pcm_default::value_type min;
      pcm_default::value_type max;
    };
    
    struct io_status
    {
      int pl_xruns;
      int ca_xruns;
    };
    
    Mixer ();
    ~Mixer ();
    
    void SetPlayback (bool state);
    void SetCapture (bool state);
    
    bool IsPlayback ();
    bool IsCapture ();
    
    track_index PlaybackChannels ();
    track_index CaptureChannels ();
    
    void StartIO (MixerSource* source, MixerParams* playback_params, MixerParams* capture_params);
    void StopIO ();
    
    /* - only (!!) for the mixer to signal the IO's end!
     * - is a must (!!) after a Node::StopIO () call ! - and is used to 
     * stop the io from the nodes side when the data is out (no loop mode)
     */ 
    void IOFinishedNotification ();
    bool IsIORunning ();
    
    // TODO: speed control (stepless fast rewind / forward ...)
    
    // for Nodes or export plugins to obtain data - size is Node's FrameSize () ...
    bool Read (track_index channel, SampleBufferDefault* destination);
    bool Write (track_index channel, SampleBufferDefault* source);
    
    void SetNode (MixerNode* new_node);
    MixerNode* GetNode ();
    
    const io_status& GetIOStatus ();
    
    inline sample_index CurrentReadPos ();
    inline sample_index CurrentWritePos ();
    
    int PlaybackPeriodSize ();
    int CapturePeriodSize ();
    
    // simple VU-Meter and over dBfs counting (I might add an plugin system for more complex stuff)
    vu_meter_data PlaybackChannelVU (sample_index channel);
    vu_meter_data CaptureChannelVU (sample_index channel);
    
    vu_meter_data PlaybackTrackVU (sample_index track);
    vu_meter_data CaptureTrackVU (sample_index track);
    
    void Debug ();
    
  protected:
    
    struct channel_period
    {
      channel_period (){
	buffer = 0;
	vu_meter = (vu_meter_data) { {0}, {0} };
	
	filled = false;
	data = false;
      }
      
      SampleBufferDefault* buffer; // only used for the channel i/o
      vu_meter_data vu_meter;
      
      bool filled; // status flag for read/wirte to check for available data / free space
      bool data; // (needs a rename?) read return value from mixer_source
    };
    
    // fill p/c vaild channels and channels
    void clear_values ();
    void build_values ();
    
    void vu_meter_from_buffer (SampleBufferDefault* buffer, vu_meter_data& data);

    void mix_next_period ();
    void write_period ();
    
    void SlotPlaybackValueChanged (track_index track, track_index channel, pcm_default::value_type value);
    Connection connection_playback_value_changed;
    void SlotCaptureValueChanged (track_index track, track_index channel, pcm_default::value_type value);
    Connection connection_capture_value_changed;
    
    void SlotPlaybackTrackMuteChanged (track_index track, bool n_state);
    Connection connection_playback_track_mute_changed;
    void SlotCaptureTrackMuteChanged (track_index track, bool n_state);
    Connection connection_capture_track_mute_changed;
    
    void SlotPlaybackChannelMuteChanged (track_index channel, bool n_state);
    Connection connection_playback_channel_mute_changed;
    void SlotCaptureChannelMuteChanged (track_index channel, bool n_state);
    Connection connection_capture_channel_mute_changed;
    
    // *** STATES ***
    bool io_running;
    
    MixerSource* m_source;
    MixerParams* m_playback_params;
    MixerParams* m_capture_params;
    
    // the nodes channels
    track_index playback_channels;
    track_index capture_channels;
    
    // the channels for which we have params
    track_index vaild_playback_channels;
    track_index vaild_capture_channels;
    
    // the tracks for which we have to pass data
    track_index vaild_playback_tracks;
    track_index vaild_capture_tracks;

    // our 2d mixer matrix [track] [channel]
    vector <vector <pcm_default::value_type> > playback_values;
    vector <vector <pcm_default::value_type> > capture_values;
    
    vector <bool> playback_track_mutes;
    vector <bool> capture_track_mutes;
    vector <bool> playback_channel_mutes;
    vector <bool> capture_channel_mutes;
    
    // io states
    int pl_period_size;
    int ca_period_size;
    
    vector <sample_index> io_read_period;
    vector <sample_index> io_write_period;
    
    // obsolete: vector <bool> io_read_pos_vaild; // read/write pos state of the source
    vector <bool> io_write_pos_vaild;
    
    SampleBufferDefault* playback_zero_buffer; // for invaild (no parameter) channels
    
    vector <channel_period> playback_channel_queues;
    vector <channel_period> capture_channel_queues;
    
    vector <channel_period> playback_track_queues;
    vector <channel_period> capture_track_queues;
    
    
    io_status m_io_status;
    
    // *** CONFIG ***
    bool io_playback;
    bool io_capture;
    
    MixerNode* m_node;
  };
  
} // end namespace GSML

#endif // GSML_IO_H
