/*
 * A Pthread wrapper based on a former implementation from the sigc++ package
 * by Karl Nelson
 *
 * (C) Valentin Ziegler 2002
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef THREAD_HH
#define THREAD_HH

// ??
#define THREAD_API

#include <pthread.h>

namespace GSMX_utils
{
  namespace Threads
  {

    struct CondAttr { pthread_condattr_t* impl_;};
    struct MutexAttr { pthread_mutexattr_t* impl_;};
    struct ThreadAttr { pthread_attr_t* impl_;};

    typedef pthread_mutex_t MutexImpl;
    typedef pthread_cond_t CondImpl;
    typedef pthread_key_t KeyImpl;
    typedef pthread_t ThreadImpl;

    // Mutual Exclusion
    class Mutex
    {
      typedef MutexImpl Impl;
    private:
      Impl mutex_;
      int destroy();

    public:
      static THREAD_API MutexAttr Default;
      operator Impl* ()  {return (Impl*)(&mutex_);}

      Mutex(const MutexAttr attr=Default); 

      // (needs work) 
      ~Mutex();

      int lock();
      int trylock();
      int unlock();
    };

    // A lazy way to unlock at end of scope
    struct MLock
    {
      Mutex &mutex_;
      MLock(Mutex& mutex):mutex_(mutex) {mutex_.lock();}
      ~MLock()                          {mutex_.unlock();}
    };

    // Condition Variable
    struct Condition
    {
      typedef CondImpl Impl;
    private:
      Impl cond_;

      int destroy();
    public:
      static THREAD_API CondAttr Default;
      operator Impl* ()  {return (Impl*)(&cond_);}

      Condition(const CondAttr &attr=Default);
      ~Condition();

      // restarts exactly one thread hung on condition
      int signal();

      // restarts all threads waiting on condition
      int broadcast();

      // unlocks a mutex while waiting on a condition, then reaquires lock.
      int wait(Mutex &m);

      // unlocks a mutex while waiting on a condition, then reaquires lock
     // with a fixed maximum duration.
      int wait(Mutex &m,struct timespec* spec);

    };

    // Integer Semaphore
    struct Semaphore
    {
      int value_;
      Condition sig_;
      Mutex access_;

      void up();
      void down();
   
      Semaphore(int value=1);
      ~Semaphore();
    };

    struct Private___
    {
      KeyImpl key_;
      void* get();
      void set(void *value);
      void create(void (*dtor)(void*));
      void destroy();
    };

    // Private__ is a thread split static.  
    template <class T>
    class Private__ : private Private___
    {
    private:
      static void dtor(void* v)
      {
	T* obj=(T*)v;
	delete obj;
      }
      
    public:

      T& operator =(const T& t)
      {return (((T&)*this)=t);}
      
      operator T& ()
      {
	T *value=(T*)get();
	if (!value)
	  set((void*)(value=new T()));  
	return *(value);
      }

      Private__()  { create(&dtor); }
      ~Private__() { destroy(); }
  };
    
    // int needs to initialized
    template <>
    class Private__<int> : private Private___
    {
    private:
      static void dtor(void* v)
        {
          int* obj=(int*)v;
          delete obj;
        }

    public:
      int& operator =(const int& t)
      {return (((int&)*this)=t);}

      operator int& ()
      {
	int *value=(int*)get();
	if (!value)
	  set((void*)(value=new int(0)));  
	return *(value); 
      }

      Private__() { create(&dtor); }
      ~Private__() { destroy(); }
    };

    struct Thread
    {
    protected:
      typedef ThreadImpl Impl;
      Impl thread_;
      void*     arg_;
      ThreadAttr attr_;
      
      static void* call_main_(void* obj);
      
    public:
      operator Impl* () {return &thread_;}

      virtual void* main(void*)=0;
      int detach();

      static THREAD_API ThreadAttr Default;

      // arg is for passing extra data to main, but never pass a
      // local variable or address of local variable.  Arg must
      // be available throughout life of program.
      int start(void* arg=0);
     
      Thread(const ThreadAttr &attr=Default);
      virtual ~Thread();
    };

  } /* namespace Threads */
} /* namespace SigC */

#endif /* THREAD_HH */
